/******************************************************************************
 * FIRESTARTER - A Processor Stress Test Utility
 * Copyright (C) 2015 TU Dresden, Center for Information Services and High
 * Performance Computing
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact: daniel.hackenberg@tu-dresden.de
 *****************************************************************************/


#include "firestarter_global.h"
#include "watchdog.h"


extern unsigned long long HIGH;

/* signal the workers:
 * 0 = use low function
 * 1 = use high function
 * 2 = exit worker as soon as one internal asm-loop is finished
 */
void set_lowhigh(unsigned long long *highvar, unsigned long long value)
{
	/* one can argue if the mfence and cpuid is necessary, but we have certain
	 * measurements which suggest that the workers reacting quicker on the
	 * shared variable change if one uses these two instructions
	 */
	__asm__ __volatile__ ("mfence;"
			      "cpuid;" ::: "eax", "ebx", "ecx", "edx");
	*highvar = value;
	__asm__ __volatile__ ("mfence;"
			      "cpuid;" ::: "eax", "ebx", "ecx", "edx");
}


/* exit with zero returncode on sigterm */
void sigterm_handler()
{
    HIGH = 2;
    usleep(5000);
    fprintf(stderr, "Caught shutdown signal, ending now ...\n");
    

    exit(EXIT_SUCCESS);
}


/* coordinates high load and low load phases
 * stops FIRESTARTER when timeout is reached
 * SPECIAL MPI Version
 */
void *watchdog_timer(watchdog_arg_t *arg)
{
	sigset_t signal_mask;
	long long timeout, time, period, load, idle, advance, load_reduction, idle_reduction;
	unsigned long long *high;
	struct timespec start_ts, current;
	int sleepret;


    sigemptyset(&signal_mask);
    sigaddset(&signal_mask, SIGINT);
    sigaddset(&signal_mask, SIGTERM);
    pthread_sigmask(SIG_BLOCK, &signal_mask, NULL);

        period = arg->period;
        load = arg->load;
        idle = period - load;
        timeout = arg->timeout;
        high = arg->high;

// Start every thread exactly synchronous

    /* signal low load */
    set_lowhigh(high, 0);

    clock_gettime(CLOCK_REALTIME, &start_ts);
    time = 0;
    /* TODO: I don't like that the control flow depends on the period variable,
     * it is not wrong but confusing
     */
    while(period > 0){
        // cycles++;


        clock_gettime(CLOCK_REALTIME, &current);
        advance = ((current.tv_sec - start_ts.tv_sec) * 1000000 + (current.tv_nsec - start_ts.tv_nsec) / 1000) % period;
        load_reduction = (load * advance) / period;
        idle_reduction = advance - load_reduction;

#ifdef ENABLE_VTRACING
        VT_USER_START("WD_HIGH");
#endif
#ifdef ENABLE_SCOREP
        SCOREP_USER_REGION_BY_NAME_BEGIN("WD_HIGH", SCOREP_USER_REGION_TYPE_COMMON);
#endif

        sleepret = usleep(load - load_reduction);
        while(sleepret != 0){ /* sometimes usleep fails, this is to be very sure it works */
            sleepret = usleep(load - load_reduction);
        }

#ifdef ENABLE_VTRACING
        VT_USER_END("WD_HIGH");
        VT_USER_START("WD_LOW");
#endif
#ifdef ENABLE_SCOREP
        SCOREP_USER_REGION_BY_NAME_END("WD_HIGH");
        SCOREP_USER_REGION_BY_NAME_BEGIN("WD_LOW", SCOREP_USER_REGION_TYPE_COMMON);
#endif

        /* signal low load */
        set_lowhigh(high, 0);
        
        sleepret = usleep(idle - idle_reduction);
        while(sleepret != 0) { /* sometimes usleep fails, this is to be very sure it works */
            sleepret = usleep(idle - idle_reduction);
        }

#ifdef ENABLE_VTRACING
        VT_USER_END("WD_LOW");
#endif
#ifdef ENABLE_SCOREP
        SCOREP_USER_REGION_BY_NAME_END("WD_LOW");
#endif

        /* signal high load */
        set_lowhigh(high, 1);
        
        time += period;
        if((timeout > 0) && (time / 1000000 >= timeout)){
            /* high == 2 signals that the workers shall shout down*/
            set_lowhigh(high, 2);
            usleep(5000);
            return 0;
        }
    }

    /* signal high load */
    set_lowhigh(high, 1);
    
    if(timeout > 0){
        sleep(timeout);
        kill(arg->pid, SIGTERM);
    }
    return 0;
}

